"""
This script calculates the physical properties of  aerosols.

The script defines constants for hygroscopic growth based on established research
and provides functions to calculate physical properties for a range of aerosol types.
It allows for customization of properties for specific aerosol categories.

"""
from typing import Dict, Any, Optional
import numpy as np

# Constants for SULFATE Hygroscopic Growth
# Fitzgerald [1975]; Haywood and Ramaswamy [1998]
SAT = np.array([0.80, 0.90, 0.95, 0.99])
F99 = 0.0155*(SAT[3] - 0.97)/(1.02 - SAT[3]**1.4)
PHI = np.array([1.058, 1.058, 1.058, 1.058-F99])
ALF = 1.2*np.exp(0.066*SAT/(PHI - SAT))
BET = np.exp(0.00077*SAT/(1.009 - SAT))
A80, A90, A95, A99 = ALF[0], ALF[1], ALF[2], ALF[3]
B80, B90, B95, B99 = BET[0], BET[1], BET[2], BET[3]

# Additional Constants
A30 = 1.0
m = (A80 - A30)/(0.80 - 0.30)
A50 = A30 + m*(0.50 - 0.30)
A70 = A30 + m*(0.70 - 0.30)
B30, B50, B70 = 1.0, 1.0, 1.0

def get_aerosol_physical_properties(
    ri_brc: Optional[float] = None,
    ri_brc_strong: Optional[float] = None,
    ri_brc_bleached: Optional[float] = None,
    ) -> Dict[str, any]:

    """
    Get the physical properties of a specific aerosols, including BrC-related types.

    Parameters:
        ri_brc (Optional[float]): Refractive index for BRC (default: None).
        ri_brc_strong (Optional[float]): Refractive index for POABRC (default: None).
        ri_brc_bleached (Optional[float]): Refractive index for SOABRC (default: None).

    Returns:
        Dict[str, Any]: Dictionary containing physical properties.

    """
    alpha_default = [1.0 for _ in range(7)]
    beta_default = [1.0 for _ in range(7)]

    base_properties = {  # rgeo, sdev, rmin, rmax, real, imag, dens, alfa[0%:99%], beta[0%:99%]
        'WATE':[1.336, 1.0e-08],
        'DUB1':[0.2986, 2.000, 0.100, 0.180, 1.531, 2.5e-03, 2.500, alpha_default,  beta_default],
        'DUB2':[0.2986, 2.000, 0.180, 0.300, 1.531, 2.5e-03, 2.500, alpha_default,  beta_default],
        'DUB3':[0.2986, 2.000, 0.300, 0.600, 1.531, 2.5e-03, 2.500, alpha_default,  beta_default],
        'DUB4':[0.2986, 2.000, 0.600, 1.000, 1.531, 2.5e-03, 2.500, alpha_default,  beta_default],
        'DUB5':[0.2986, 2.000, 1.000, 1.800, 1.531, 2.5e-03, 2.650, alpha_default,  beta_default],
        'DUB6':[0.2986, 2.000, 1.800, 3.000, 1.531, 2.5e-03, 2.650, alpha_default,  beta_default],
        'DUB7':[0.2986, 2.000, 3.000, 6.000, 1.531, 2.5e-03, 2.650, alpha_default,  beta_default],
        'DUB8':[0.2986, 2.000, 6.000, 10.00, 1.531, 2.5e-03, 2.650, alpha_default,  beta_default],
        'SSB1':[0.1500, 2.800, 0.100, 0.180, 1.557, 1.0e-08, 2.160, [1.0,1.6,1.8,2.0,2.4,2.9,4.8], beta_default],
        'SSB2':[0.1500, 2.800, 0.180, 0.300, 1.557, 1.0e-08, 2.160, [1.0,1.6,1.8,2.0,2.4,2.9,4.8], beta_default],
        'SSB3':[0.1500, 2.800, 0.300, 0.600, 1.557, 1.0e-08, 2.160, [1.0,1.6,1.8,2.0,2.4,2.9,4.8], beta_default],
        'SSB4':[0.1500, 2.800, 0.600, 1.000, 1.557, 1.0e-08, 2.160, [1.0,1.6,1.8,2.0,2.4,2.9,4.8], beta_default],
        'SSB5':[0.1500, 2.800, 1.000, 1.800, 1.557, 1.0e-08, 2.160, [1.0,1.6,1.8,2.0,2.4,2.9,4.8], beta_default],
        'SSB6':[0.1500, 2.800, 1.800, 3.000, 1.557, 1.0e-08, 2.160, [1.0,1.6,1.8,2.0,2.4,2.9,4.8], beta_default],
        'SSB7':[0.1500, 2.800, 3.000, 6.000, 1.557, 1.0e-08, 2.160, [1.0,1.6,1.8,2.0,2.4,2.9,4.8], beta_default],
        'SSB8':[0.1500, 2.800, 6.000, 15.00, 1.557, 1.0e-08, 2.160, [1.0,1.6,1.8,2.0,2.4,2.9,4.8], beta_default],
        'POA0':[0.0212, 2.200, 0.005, 20.00, 1.501, 1.0e-08, 1.800, alpha_default,  beta_default],
        'POA1':[0.0212, 2.200, 0.005, 20.00, 1.501, 1.0e-04, 1.800, [1.0,1.2,1.4,1.5,1.6,1.8,2.2], beta_default],
        'POW0':[0.0212, 2.200, 0.005, 20.00, 1.501, 1.0e-01, 1.800, alpha_default,  beta_default],
        'POW1':[0.0212, 2.200, 0.005, 20.00, 1.501, 1.0e-03, 1.800, [1.0,1.2,1.4,1.5,1.6,1.8,2.2], beta_default],
        'SOB1':[0.0212, 2.200, 0.005, 20.00, 1.486, 2.5e-05, 1.800, [1.0,1.2,1.4,1.5,1.6,1.8,2.2], beta_default],
        'SOB5':[0.0212, 2.200, 0.005, 20.00, 1.486, 2.5e-05, 1.800, [1.0,1.2,1.4,1.5,1.6,1.8,2.2], beta_default],
        'SOB6':[0.0212, 2.200, 0.005, 20.00, 1.486, 2.5e-05, 1.800, [1.0,1.2,1.4,1.5,1.6,1.8,2.2], beta_default],
        'SOB7':[0.0212, 2.200, 0.005, 20.00, 1.486, 2.5e-05, 1.800, [1.0,1.2,1.4,1.5,1.6,1.8,2.2], beta_default],
        'SUL1':[0.0695, 2.030, 0.005, 20.00, 1.546, 1.0e-08, 1.700, [A30,A50,A70,A80,A90,A95,A99], [B30,B50,B70,B80,B90,B95,B99]],
        'DUPM10':[0.2986, 2.000, 0.100, 5.00, 1.531, 2.5e-03, 2.650, alpha_default,  beta_default],
        'SSPM10':[0.1500, 2.800, 0.100, 5.00, 1.557, 1.0e-08, 2.160, [1.0,1.6,1.8,2.0,2.4,2.9,4.8], beta_default],
        'EPOA':[0.0212, 2.200, 0.005, 20.00, 1.501, 1.0e-08, 1.800, alpha_default,  beta_default],
        'OPOA':[0.0212, 2.200, 0.005, 20.00, 1.501, 1.0e-04, 1.800, [1.0,1.2,1.4,1.5,1.6,1.8,2.2], beta_default],
        'ASOA':[0.0212, 2.200, 0.005, 20.00, 1.486, 2.5e-05, 1.800, [1.0,1.2,1.4,1.5,1.6,1.8,2.2], beta_default],
        'TSOA':[0.0212, 2.200, 0.005, 20.00, 1.486, 2.5e-05, 1.800, [1.0,1.2,1.4,1.5,1.6,1.8,2.2], beta_default],
        'ISOA':[0.0212, 2.200, 0.005, 20.00, 1.486, 2.5e-05, 1.800, [1.0,1.2,1.4,1.5,1.6,1.8,2.2], beta_default],
        'FSOA':[0.0212, 2.200, 0.005, 20.00, 1.486, 2.5e-05, 1.800, [1.0,1.2,1.4,1.5,1.6,1.8,2.2], beta_default],
        'NIPM10':[0.0695, 2.030, 0.005, 20.00, 1.546, 1.0e-08, 1.700, [A30,A50,A70,A80,A90,A95,A99], [B30,B50,B70,B80,B90,B95,B99]],
        'NH1':[0.0695, 2.030, 0.005, 20.00, 1.546, 1.0e-08, 1.700, [A30,A50,A70,A80,A90,A95,A99], [B30,B50,B70,B80,B90,B95,B99]],
        'SUL1':[0.0695, 2.030, 0.005, 20.00, 1.546, 1.0e-08, 1.700, [A30,A50,A70,A80,A90,A95,A99], [B30,B50,B70,B80,B90,B95,B99]],
        'UNSPM10':[0.2986, 2.000, 0.100, 5.00, 1.531, 2.5e-03, 2.650, alpha_default,  beta_default],
        'BCPHOB':[0.0118, 2.000, 0.005, 20.00, 1.850, 7.1e-01, 1.000, alpha_default,  beta_default], #phob
        'BCPHIL':[0.0118, 2.000, 0.005, 20.00, 1.850, 7.1e-01, 1.000, [1.0,1.0,1.0,1.2,1.4,1.5,1.9], beta_default], #philic
        'POA':[0.0212, 2.200, 0.005, 20.00, 1.501, 5.7e-3, 1.800, alpha_default,  beta_default], #phob
        'OPOA':[0.0212, 2.200, 0.005, 20.00, 1.501, 1.0e-04, 1.800, [1.0,1.2,1.4,1.5,1.6,1.8,2.2], beta_default], #philic
        'SOA':[0.0212, 2.200, 0.005, 20.00, 1.486, 2.5e-05, 1.800, [1.0,1.2,1.4,1.5,1.6,1.8,2.2], beta_default],
        'BRC':[0.0212, 2.200, 0.005, 20.00, 1.501, ri_brc, 1.800, alpha_default,  beta_default], #phobic
        'POABRC':[0.0212, 2.200, 0.005, 20.00, 1.501, ri_brc_strong, 1.800, alpha_default,  beta_default],
        'SOABRC':[0.0212, 2.200, 0.005, 20.00, 1.486, ri_brc_bleached, 1.800, [1.0,1.2,1.4,1.5,1.6,1.8,2.2], beta_default]

    }
    return base_properties
